/*
	Copyright 1999, Be Incorporated.   All Rights Reserved.
	This file may be used under the terms of the Be Sample Code License.

	SiS6326 graphics card accelerant
	written by Yasutaka Uematu (yasutaka2000@bemail.org)
*/

#include	<support/Debug.h>	// y_SiS

#include "GlobalData.h"
#include "generic.h"
#include <unistd.h>
#include	"SiSaccelData.h"
#include	"SiSaccelutil.h"
#include	"IOaccess.h"	// y_SiS

extern void SiSInitRes(display_mode *dm);

uint32 calcBitsPerPixel(uint32);
uint8	SiSGetSR11(void);

/*
	Enable/Disable interrupts.  Just a wrapper around the
	ioctl() to the kernel driver.
*/
static void interrupt_enable(bool flag) {
	status_t result;
	sample_set_bool_state sbs;

	SERIAL_PRINT(("ySiS accel : enter interrupt_enable %d\n", flag));	// y_SiS
	/* set the magic number so the driver knows we're for real */
	sbs.magic = YSIS_PRIVATE_DATA_MAGIC;
	sbs.do_it = flag;
	/* contact driver and get a pointer to the registers and shared data */
	result = ioctl(fd, YSIS_RUN_INTERRUPTS, &sbs, sizeof(sbs));
}

/*
	Calculates the number of bits for a given color_space.
	Usefull for mode setup routines, etc.
*/
uint32 calcBitsPerPixel(uint32 cs) {
	uint32	bpp = 0;

	switch (cs) {
		case B_RGB32_BIG:
		case B_RGBA32_BIG:
		case B_RGB32_LITTLE:
		case B_RGBA32_LITTLE:
			bpp = 32; break;
		case B_RGB24_BIG:
		case B_RGB24_LITTLE:
			bpp = 24; break;
		case B_RGB16_BIG:
		case B_RGB16_LITTLE:
			bpp = 16; break;
		case B_RGB15_BIG:
		case B_RGBA15_BIG:
		case B_RGB15_LITTLE:
		case B_RGBA15_LITTLE:
			bpp = 15; break;
		case B_CMAP8:
			bpp = 8; break;
	}
	return bpp;
}

/*
	The code to actually configure the display.  Unfortunately, there's not much
	that can be provided in the way of sample code.  If you're lucky, you're writing
	a driver for a device that has all (or at least most) of the bits for a particular
	configuration value in the same register, rather than spread out over the standard
	VGA registers + a zillion expansion bits.  In any case, I've found that the way
	to simplify this routine is to do all of the error checking in PROPOSE_DISPLAY_MODE(),
	and just assume that the values I get here are acceptable.
*/
static void do_set_display_mode(display_mode *dm) {
	/* disable interrupts using the kernel driver */
	interrupt_enable(false);
	SiSInitRes(dm);		// y_SiS
	/* enable interrupts using the kernel driver */
	interrupt_enable(true);
}

/*
	The exported mode setting routine.  First validate the mode, then call our
	private routine to hammer the registers.
*/
status_t SET_DISPLAY_MODE(display_mode *mode_to_set) {
//	display_mode bounds, target;
	static display_mode bounds, target;	// BUG? 001028 TEST Cheap workaround

	SERIAL_PRINT(("ySiS accel : enter SET_DISPLAY_MODE\n"));	// TEST 001028
	
	/* ask for the specific mode */
	target = bounds = *mode_to_set;
	if (PROPOSE_DISPLAY_MODE(&target, &bounds, &bounds) == B_ERROR)
		return B_ERROR;
	do_set_display_mode(&target);
	si->dm = target;	// y_SiS
	si->fbc.bytes_per_row = ((calcBitsPerPixel(si->dm.space) / 8) * si->dm.virtual_width);	// y_SiS
	
	SERIAL_PRINT(("ySiS accel : leave SET_DISPLAY_MODE\n"));	// TEST 001028
	return B_OK;
}

/*
	Set which pixel of the virtual frame buffer will show up in the
	top left corner of the display device.  Used for page-flipping
	games and virtual desktops.
*/
status_t MOVE_DISPLAY(uint16 h_display_start, uint16 v_display_start) {
	union {
		uint32	fAddr;
		uint8	fEAddr[4];
	} u;
	SERIAL_PRINT(("ySiS accel : enter MOVE_DISPLAY()\n"));	// y_SiS
	/*
	Many devices have limitations on the granularity of the horizontal offset.
	Make any checks for this here.  A future revision of the driver API will
	add a hook to return the granularity for a given display mode.
	*/
	/* most cards can handle multiples of 8 */
	// ySiS : SiS6326は、8bppではX方向4pixel単位、16bppではX方向2pixel単位の制限がある。
	switch(si->dm.space) {
	case B_CMAP8:
		if (h_display_start & 0x03) {
			return B_ERROR;
		}
		u.fAddr = (v_display_start * si->fbc.bytes_per_row / 4) + h_display_start / 4;
		break;
	case B_RGB15_BIG:	// y_SiS ??? どちらか？
	case B_RGBA15_BIG:	// y_SiS ???
	case B_RGB15_LITTLE:	// y_SiS ???
	case B_RGBA15_LITTLE:	// y_SiS ???
	case B_RGB16_BIG:
	case B_RGB16_LITTLE:
		if (h_display_start & 0x01) {
			return B_ERROR;
		}
		u.fAddr = (v_display_start * si->fbc.bytes_per_row / 2) + h_display_start / 2;
		break;
	}
	/* do not run past end of display */
	if ((si->dm.timing.h_display + h_display_start) > si->dm.virtual_width) {
		SERIAL_PRINT(("ySiS accel : MOVE_DISPLAY ERROR 1\n"));
		return B_ERROR;
	}
	if ((si->dm.timing.v_display + v_display_start) > si->dm.virtual_height) {
		SERIAL_PRINT(("ySiS accel : MOVE_DISPLAY ERROR 2\n"));
		return B_ERROR;
	}

	/* everybody remember where we parked... */
	si->dm.h_display_start = h_display_start;
	si->dm.v_display_start = v_display_start;

//	SERIAL_PRINT(("ySiS accel : MOVE_DISPLAY to %08x\n", u.fAddr));
	/* actually set the registers */
	gSEQR.s.fSR30.FPFSaddress70 = u.fEAddr[3];
	gSEQR.s.fSR31.FPFSaddress158 = u.fEAddr[2];
	gSEQR.s.fSR32.FPFSaddress1916 = u.fEAddr[1];
	SeqOut(0x30, gSEQR.fValue[0x30]);
	SeqOut(0x31, gSEQR.fValue[0x31]);
	SeqOut(0x32, gSEQR.fValue[0x32]);
	
	return B_OK;
}

/*
	Set the indexed color palette.

	このフックは、垂直帰線割り込みが禁止された状態で呼ばれる。
*/
void SET_INDEXED_COLORS(uint count, uint8 first, uint8 *color_data, uint32 flags) {
	uint8	theValue;		// y_SiS
	/*
	Some cards use the indexed color regisers in the DAC for gamma correction.
	If this is true with your device (and it probably is), you need to protect
	against setting these registers when not in an indexed mode.
	*/
	SERIAL_PRINT(("ySiS accel : enter SET_INDEXED_COLORS()\n"));	// y_SiS
	if (si->dm.space != B_CMAP8) return;

	/*
	There isn't any need to keep a copy of the data being stored, as the app_server
	will set the colors each time it switches to an 8bpp mode, or takes ownership
	of an 8bpp mode after a GameKit app has used it.
	*/

	/*
	You're on your own from here.  Chances are your code will look something like
	what follows:
	*/
	outb(DAC_WRITE_IDX, first);
	while(count--) {
		theValue = *color_data++;
		outb(DAC_DATA, (uint8)(theValue >> 2));	// R
		theValue = *color_data++;
		outb(DAC_DATA, (uint8)(theValue >> 2));	// G
		theValue = *color_data++;
		outb(DAC_DATA, (uint8)(theValue >> 2));	// B
	}
}


/* masks for DPMS control bits */
enum {
	H_SYNC_OFF = 0x80,
	V_SYNC_OFF = 0x40,
	BITSMASK = H_SYNC_OFF | V_SYNC_OFF
};

/*
	Put the display into one of the Display Power Management modes.
*/
status_t SET_DPMS_MODE(uint32 dpms_flags) {
	uint8 LTemp;

	SERIAL_PRINT(("ySiS accel : enter SET_DPMS_MODE %08x\n", dpms_flags));	// ySiS
	/*
	The register containing the horizontal and vertical sync control bits
	usually contains other control bits, so do a read-modify-write.
	*/
	LTemp = SiSGetSR11();
		
	/* this mask is device dependant */
	LTemp &= ~BITSMASK;	/* clear all disable bits (including display disable) */

	/* now pick one of the DPMS configurations */
	switch(dpms_flags) {
		case B_DPMS_ON:	/* H: on, V: on */
			/* do nothing, bits already clear */
			/* usually, but it may be different for your device */
			break;
		case B_DPMS_STAND_BY: /* H: on, V: off */
			LTemp |= V_SYNC_OFF;
			break;
		case B_DPMS_SUSPEND: /* H: off, V: on */
			LTemp |= H_SYNC_OFF;
			break;
		case B_DPMS_OFF: /* H: off, V: off, display off */
			LTemp |= H_SYNC_OFF | V_SYNC_OFF;
			break;
		default:
			return B_ERROR;
	}
	{
		tIOIdxAccess	theSR11;

		theSR11.fOffset = SEQ_IDX;
		theSR11.fIndex = 0x11;
		theSR11.fValue = LTemp;
		(void)ioctl(fd, YSIS_WRITE_INDEXED, &theSR11);
		gSEQR.fValue[0x11] = theSR11.fValue;
	}

	/*
	NOTE: if you're driving a device with a backlight (like a digital LCD),
	you may want to turn off the backlight here when in a DPMS power saving mode.
	*/
	if (dpms_flags == B_DPMS_ON)
		/* turn on the back light */
		;
	else
		/* turn off the back light */
		;

	SERIAL_PRINT(("ySiS accel : leave SET_DPMS_MODE\n"));	// ySiS
	return B_OK;
}

/*
	Report device DPMS capabilities.  Most newer cards can do it all.
	I've only seen one older card that can do a subset (ON and OFF).
	Very early cards may not be able to do this at all.
*/
uint32 DPMS_CAPABILITIES(void) {
	SERIAL_PRINT(("ySiS accel : enter - leave DPMS_CAPABILITIES\n"));	// ySiS
	return 	B_DPMS_ON | B_DPMS_STAND_BY  | B_DPMS_SUSPEND | B_DPMS_OFF;
}


/*
	Return the current DPMS mode.
*/
uint32 DPMS_MODE(void) {
	uint8 LTemp;
	uint32 mode = B_DPMS_ON;
	
	SERIAL_PRINT(("ySiS accel : enter DPMS_MODE\n"));	// ySiS
	/* read the control bits from the device */
	LTemp = SiSGetSR11();
		
	/* what mode is set? */
	switch (LTemp & (H_SYNC_OFF | V_SYNC_OFF)) {
		case 0:	/* H: on, V: on */
			mode = B_DPMS_ON;
			break;
		case H_SYNC_OFF: /* H: off, V: on */
			mode = B_DPMS_STAND_BY;
			break;
		case V_SYNC_OFF: /* H: on, V: off */
			mode = B_DPMS_SUSPEND;
			break;
		case (H_SYNC_OFF | V_SYNC_OFF): /* H: off, V: off */
			mode = B_DPMS_OFF;
			break;
	}
	SERIAL_PRINT(("ySiS accel : leave DPMS_MODE %08x\n", mode));	// ySiS
	return mode;
}

uint8 SiSGetSR11(void) {
	tIOIdxAccess	theSR11;

	theSR11.fOffset = SEQ_IDX;
	theSR11.fIndex = 0x11;
	(void)ioctl(fd, YSIS_READ_INDEXED, &theSR11);
	gSEQR.fValue[0x11] = theSR11.fValue;

	return theSR11.fValue;
}
